"""Visible volume code."""

import functools
from typing import Any, Iterable, Optional, Union

import numpy as np

from morfeus.geometry import Cone
from morfeus.io import read_geometry
from morfeus.sasa import SASA
from morfeus.typing import ArrayLike1D, ArrayLike2D
from morfeus.utils import check_distances, convert_elements, get_radii


class VisibleVolume:
    """Calculates and stores visible volume and area.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        metal_index: Index of metal atom (1-indexed)
        include_hs: Whether to include H atoms in the calculation
        radii: Atomic radii (Å)
        radii_type: Radii type: 'alvarez', 'bondi', 'crc', 'pyykko', 'rahm' or 'truhlar'
        radius: Radius of sphere to divide proximal and distal (Å)
        density: Area per point on atom surface (Å²)

    Attributes:
        distal_area: Distal area (Å²)
        distal_visible_area: Distal visible area (Å²)
        distal_visible_volume: Distal visible volume (Å³)
        distal_volume: Distal volume (Å³)
        invisible_area: Invisible area (Å²)
        invisible_volume: Invisible volume (Å³)
        proximal_area: Proximal area (Å²)
        proximal_visible_area: Proximal visible area (Å²)
        proximal_visible_volume: Proximal visible volume (Å³)
        proximal_volume: Proximal volume (Å³)
        total_area: Total area (Å²)
        total_volume: Total volume (Å³)
        visible_area: Visible area (Å²)
        visible_volume: Visible volume (Å³)
    """

    distal_area: float
    distal_visible_area: float
    distal_visible_volume: float
    distal_volume: float
    invisible_area: float
    invisible_volume: float
    proximal_area: float
    proximal_visible_area: float
    proximal_visible_volume: float
    proximal_volume: float
    total_area: float
    total_volume: float
    visible_area: float
    visible_volume: float

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        metal_index: int,
        include_hs: bool = True,
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "pyykko",
        radius: float = 3.5,
        density: float = 0.01,
    ) -> None:
        # Set up arrays and get radii
        elements = np.array(convert_elements(elements, output="numbers"))
        coordinates = np.array(coordinates)
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)
        radii = np.array(radii)

        # Check so that no atom is within vdW distance of metal atom
        within = check_distances(elements, coordinates, metal_index, radii=radii)
        if len(within) > 0:
            atom_string = " ".join([str(i) for i in within])
            raise Exception("Atoms within vdW radius of central atom:", atom_string)

        # Center coordinate system around metal and remove it
        coordinates -= coordinates[metal_index - 1]
        elements = np.delete(elements, metal_index - 1)
        coordinates = np.delete(coordinates, metal_index - 1, axis=0)
        radii = np.delete(radii, metal_index - 1)

        # Remove H atoms
        if include_hs is False:
            h_mask = elements == 1
            elements = elements[~h_mask]
            coordinates = coordinates[~h_mask]
            radii = radii[~h_mask]

        # Construct SASA object
        sasa = SASA(
            elements,
            coordinates,
            radii=radii,
            radii_type=radii_type,
            probe_radius=0.0,
            density=density,
        )

        # Set invisible and proximal masks for each atom
        atoms = sasa._atoms
        for atom in atoms:
            atom.invisible_mask = np.zeros(len(atom.accessible_points), dtype=bool)
            atom.proximal_mask = np.linalg.norm(atom.accessible_points, axis=1) < radius

        # Check points on other atoms against cone for each atom
        atom_coordinates = np.array([atom.coordinates for atom in atoms])
        atoms = np.array(atoms)
        for atom in atoms:
            # Calculate distances to other atoms
            atom.get_cone()
            cone = Cone(atom.cone.angle, [atom.index], atom.cone.normal)
            atom_dist = np.linalg.norm(atom.coordinates)
            other_distances = np.dot(atom_coordinates, cone.normal)

            # Check whether points are (1) within cone and (2) beyond atom
            # center
            check_atoms = atoms[other_distances >= (atom_dist - atom.radius)]
            for check_atom in check_atoms:
                if check_atom == atom:
                    continue
                is_inside_mask = cone.is_inside_points(
                    check_atom.accessible_points, method="cross"
                )
                is_beyond_mask = (
                    np.dot(check_atom.accessible_points, cone.normal) >= atom_dist
                )
                invisible_mask = np.logical_and(is_inside_mask, is_beyond_mask)
                check_atom.invisible_mask = np.logical_or(
                    check_atom.invisible_mask, invisible_mask
                )

        # Calculate visible, invisible and proximal_visible volume
        visible_volume = 0
        invisible_volume = 0
        proximal_visible_volume = 0
        proximal_volume = 0
        visible_area = 0
        invisible_area = 0
        proximal_visible_area = 0
        proximal_area = 0
        for atom in atoms:
            point_areas = atom.point_areas[atom.accessible_mask]
            point_volumes = atom.point_volumes[atom.accessible_mask]

            invisible_volume += point_volumes[atom.invisible_mask].sum()
            visible_volume += point_volumes[~atom.invisible_mask].sum()
            proximal_visible_volume += point_volumes[
                np.logical_and(~atom.invisible_mask, atom.proximal_mask)
            ].sum()
            proximal_volume += point_volumes[atom.proximal_mask].sum()

            invisible_area += point_areas[atom.invisible_mask].sum()
            visible_area += point_areas[~atom.invisible_mask].sum()
            proximal_visible_area += point_areas[
                np.logical_and(~atom.invisible_mask, atom.proximal_mask)
            ].sum()
            proximal_area += point_areas[atom.proximal_mask].sum()

        # Store attributes
        self.total_volume = sasa.volume
        self.proximal_volume = proximal_volume
        self.distal_volume = sasa.volume - proximal_volume
        self.invisible_volume = invisible_volume
        self.visible_volume = visible_volume
        self.proximal_visible_volume = proximal_visible_volume
        self.distal_visible_volume = visible_volume - proximal_visible_volume

        self.total_area = sasa.area
        self.proximal_area = proximal_area
        self.distal_area = sasa.area - proximal_area
        self.invisible_area = invisible_area
        self.visible_area = visible_area
        self.proximal_visible_area = proximal_visible_area
        self.distal_visible_area = visible_area - proximal_visible_area

        self._atoms = atoms

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for visible volume.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(VisibleVolume, elements, coordinates)



"""Run in command line mode."""
import fire

import morfeus.buried_volume
import morfeus.cone_angle
import morfeus.conformer
import morfeus.dispersion
import morfeus.local_force
import morfeus.pyramidalization
import morfeus.sasa
import morfeus.sterimol
import morfeus.visible_volume
import morfeus.xtb


def main() -> None:
    """Call Fire to access command line scripts."""
    fire.Fire(
        {
            "buried_volume": morfeus.buried_volume.cli,
            "cone_angle": morfeus.cone_angle.cli,
            "conformer": morfeus.conformer.cli,
            "dispersion": morfeus.dispersion.cli,
            "local_force": morfeus.local_force.cli,
            "pyramidalization": morfeus.pyramidalization.cli,
            "sasa": morfeus.sasa.cli,
            "sterimol": morfeus.sterimol.cli,
            "visible_volume": morfeus.visible_volume.cli,
            "xtb": morfeus.xtb.cli,
        }
    )


if __name__ == "__main__":
    main()



"""Solvent accessible surface area code."""

import functools
import typing
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import scipy.spatial

from morfeus.data import atomic_symbols, jmol_colors
from morfeus.geometry import Atom, Sphere
from morfeus.io import read_geometry
from morfeus.typing import ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import pyvista as pv
    from pyvistaqt import BackgroundPlotter


class SASA:
    """Performs and stores results of solvent accessible surface area calculations.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        radii: VdW radii (Å)
        radii_type: Choice of vdW radii: 'bondi' or 'crc' (default)
        probe_radius: Radius of probe atom (Å)
        density: Area per point (Å²) on the vdW surface

    Attributes:
        area: Area of the solvent accessible surface.
        atom_areas: Atom areas (starting from 1)
        atom_volumes: Atom volumes (starting from 1)
        volume: Volume of the solvent accessible surface
    """

    area: float
    atom_areas: Dict[int, float]
    atom_volumes: Dict[int, float]
    volume: float
    _atoms: List[Atom]
    _density: float
    _probe_radius: float

    def __init__(
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "crc",
        probe_radius: float = 1.4,
        density: float = 0.01,
    ) -> None:
        # Converting elements to atomic numbers if the are symbols
        elements = convert_elements(elements, output="numbers")
        coordinates = np.array(coordinates)

        # Getting radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)

        # Increment the radii with the probe radius
        radii = np.array(radii)
        radii = radii + probe_radius

        # Construct list of atoms
        atoms = []
        for i, (coordinate, radius, element) in enumerate(
            zip(coordinates, radii, elements), start=1
        ):
            atom = Atom(element, coordinate, radius, i)
            atoms.append(atom)

        # Set up attributes
        self._atoms = atoms
        self._density = density
        self._probe_radius = probe_radius

        # Determine accessible and occluded points for each atom
        self._determine_accessible_points()

        # Calculate atom areas and volumes
        self._calculate()

    def _calculate(self) -> None:
        """Calculate solvent accessible surface area and volume."""
        for atom in self._atoms:
            # Get number of points of eache type
            n_accessible = len(atom.accessible_points)
            n_occluded = len(atom.occluded_points)
            n_points = len(atom.accessible_points) + len(atom.occluded_points)

            # Calculate part occluded and accessible
            ratio_occluded = n_occluded / n_points
            ratio_accessible = 1 - ratio_occluded

            # Calculate area
            area = 4 * np.pi * atom.radius ** 2 * ratio_accessible
            atom.area = area
            atom.point_areas = np.zeros(n_points)
            if n_accessible > 0:
                atom.point_areas[atom.accessible_mask] = atom.area / n_accessible

            # Center accessible points and normalize
            centered_points = np.array(atom.accessible_points) - atom.coordinates
            centered_points /= np.linalg.norm(centered_points, axis=1).reshape(-1, 1)

            # Add accessible points
            accessible_summed = np.sum(centered_points, axis=0)

            # Calculate volume
            volume = (4 * np.pi / 3 / n_points) * (
                atom.radius ** 2 * np.dot(atom.coordinates, accessible_summed)
                + atom.radius ** 3 * n_accessible
            )
            atom.volume = volume
            atom.point_volumes = np.zeros(n_points)
            if n_accessible > 0:
                atom.point_volumes[atom.accessible_mask] = atom.volume / n_accessible

        # Set up attributes
        self.atom_areas = {atom.index: atom.area for atom in self._atoms}
        self.atom_volumes = {atom.index: atom.volume for atom in self._atoms}
        self.area = sum([atom.area for atom in self._atoms])
        self.volume = sum([atom.volume for atom in self._atoms])

    def _determine_accessible_points(self) -> None:
        """Determine occluded and accessible points of each atom."""
        # Based on distances to all other atoms (brute force).
        for atom in self._atoms:
            # Construct sphere for atom
            sphere = Sphere(atom.coordinates, atom.radius, density=self._density)
            atom.points = sphere.points

            # Select atoms that are at a distance less than the sum of radii
            # !TODO can be vectorized
            test_atoms = []
            for test_atom in self._atoms:
                if test_atom is not atom:
                    distance = scipy.spatial.distance.euclidean(
                        atom.coordinates, test_atom.coordinates
                    )
                    radii_sum = atom.radius + test_atom.radius
                    if distance < radii_sum:
                        test_atoms.append(test_atom)

            # Select coordinates and radii for other atoms
            test_coordinates = [test_atom.coordinates for test_atom in test_atoms]
            test_radii = [test_atom.radius for test_atom in test_atoms]
            test_radii = np.array(test_radii).reshape(-1, 1)

            # Get distances to other atoms and subtract radii
            if test_coordinates:
                distances = scipy.spatial.distance.cdist(
                    test_coordinates, sphere.points
                )
                distances -= test_radii
                # Take smallest distance and perform check
                min_distances = np.min(distances, axis=0)
                atom.occluded_mask = min_distances < 0
                atom.accessible_mask = ~atom.occluded_mask
            else:
                atom.occluded_mask = np.zeros(len(atom.points), dtype=bool)
                atom.accessible_mask = np.ones(len(atom.points), dtype=bool)
            atom.occluded_points = sphere.points[atom.occluded_mask]
            atom.accessible_points = sphere.points[atom.accessible_mask]

    def print_report(self, verbose: bool = False) -> None:
        """Print report of results.

        Args:
            verbose: Whether to print atom areas
        """
        print(f"Probe radius (Å): {self._probe_radius}")
        print(f"Solvent accessible surface area (Å²): {self.area:.1f}")
        print("Volume inside solvent accessible surface (Å³): " f"{self.volume:.1f}")
        if verbose:
            print(f"{'Symbol':<10s}{'Index':<10s}{'Area (Å²)':<10s}")
            for atom, (i, area) in zip(self._atoms, self.atom_areas.items()):
                symbol = atomic_symbols[atom.element]
                print(f"{symbol:<10s}{i:<10d}{area:<10.1f}")

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        atom_scale: float = 1,
        background_color: str = "white",
        point_color: str = "steelblue",
        opacity: float = 0.25,
        size: float = 1,
    ) -> None:
        """Draw a 3D representation.

        Draws the molecule with the solvent accessible surface area.

        Args:
            atom_scale: Scaling factor for atom size
            background_color: Background color for plot
            point_color: Color of surface points
            opacity: Point opacity
            size: Point size
        """
        # Set up plotter
        p = BackgroundPlotter()
        p.set_background(background_color)

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            radius = atom.radius * atom_scale - self._probe_radius
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            p.add_mesh(sphere, color=color, opacity=1, name=str(atom.index))

        # Draw surface points
        surface_points = np.vstack([atom.accessible_points for atom in self._atoms])
        p.add_points(
            surface_points, color=point_color, opacity=opacity, point_size=size
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for solvent accessible surface area.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(SASA, elements, coordinates)



"""Cone angle code."""

import functools
import itertools
import math
import typing
from typing import Any, Iterable, List, Optional, Set, Union

import numpy as np

from morfeus.data import atomic_symbols, jmol_colors
from morfeus.geometry import Atom, Cone
from morfeus.io import read_geometry
from morfeus.plotting import get_drawing_cone
from morfeus.typing import ArrayLike1D, ArrayLike2D
from morfeus.utils import (
    check_distances,
    convert_elements,
    get_radii,
    Import,
    requires_dependency,
)

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import pyvista as pv
    from pyvistaqt import BackgroundPlotter


class ConeAngle:
    """Calculates and stores the results of exact cone angle calculation.

    As described in J. Comput. Chem. 2013, 34, 1189.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        atom_1: Index of central atom (1-inexed)
        radii: vdW radii (Å)
        radii_type: Type of vdW radii: 'alvarez', 'bondi', 'crc' or 'truhlar'

    Attributes:
        cone_angle: Exact cone angle (degrees)
        tangent_atoms: Atoms tangent to cone (1-indexed)

    Raises:
        Exception: If cone angle could not be found
    """

    cone_angle: float
    tangent_atoms: List[int]
    _atoms: List[Atom]
    _max_2_cone: Cone

    def __init__(  # noqa: C901
        self,
        elements: Union[Iterable[int], Iterable[str]],
        coordinates: ArrayLike2D,
        atom_1: int,
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "crc",
    ) -> None:
        # Convert elements to atomic numbers if the are symbols
        elements = convert_elements(elements, output="numbers")
        coordinates = np.array(coordinates)

        # Get radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)
        radii = np.array(radii)

        # Check so that no atom is within vdW distance of atom 1
        within = check_distances(elements, coordinates, atom_1, radii=radii)
        if len(within) > 0:
            atom_string = " ".join([str(i) for i in within])
            raise Exception("Atoms within vdW radius of central atom:", atom_string)

        # Set up coordinate array and translate coordinates
        coordinates -= coordinates[atom_1 - 1]

        # Get list of atoms as Atom objects
        atoms: List[Atom] = []
        for i, (element, coord, radius) in enumerate(
            zip(elements, coordinates, radii), start=1
        ):
            if i != atom_1:
                atom = Atom(element, coord, radius, i)
                atom.get_cone()
                atoms.append(atom)
        self._atoms = atoms

        # Search for cone over single atoms
        cone = self._search_one_cones()

        # Prune out atoms that lie in the shadow of another atom's cone
        if cone is None:
            loop_atoms = list(atoms)
            remove_atoms: Set[Atom] = set()
            for cone_atom in loop_atoms:
                for test_atom in loop_atoms:
                    if cone_atom is not test_atom:
                        if cone_atom.cone.is_inside(test_atom):
                            remove_atoms.add(test_atom)
            for atom in remove_atoms:
                loop_atoms.remove(atom)
            self._loop_atoms = loop_atoms

        # Search for cone over pairs of atoms
        if cone is None:
            cone = self._search_two_cones()

        # Search for cones over triples of atoms
        if cone is None:
            cone = self._search_three_cones()

        # Set attributes
        self._cone = cone
        self.cone_angle = math.degrees(cone.angle * 2)
        self.tangent_atoms = [atom.index for atom in cone.atoms]

    def print_report(self) -> None:
        """Prints report of results."""
        tangent_atoms = [
            atom for atom in self._atoms if atom.index in self.tangent_atoms
        ]
        tangent_labels = [
            f"{atomic_symbols[atom.element]}{atom.index}" for atom in tangent_atoms
        ]
        tangent_string = " ".join(tangent_labels)
        print(f"Cone angle: {self.cone_angle:.1f}")
        print(f"No. tangent atoms: {len(tangent_atoms)}")
        print(f"Tangent to: {tangent_string}")

    def _get_upper_bound(self) -> float:
        """Calculates upper bound for apex angle.

        Returns:
            upper_bound: Upper bound to apex angle (radians)
        """
        # Calculate unit vector to centroid
        coordinates = np.array([atom.coordinates for atom in self._atoms])
        centroid_vector = np.mean(coordinates, axis=0)
        centroid_unit_vector = centroid_vector / np.linalg.norm(centroid_vector)

        # Getting sums of angle to centroid and vertex angle.
        angle_sums = []
        for atom in self._atoms:
            cone = atom.cone
            cos_angle = np.dot(centroid_unit_vector, cone.normal)
            vertex_angle = math.acos(cos_angle)
            angle_sum = cone.angle + vertex_angle
            angle_sums.append(angle_sum)

        # Select upper bound as the maximum angle
        upper_bound = max(angle_sums)

        return upper_bound

    def _search_one_cones(self) -> Optional[Cone]:
        """Searches over cones tangent to one atom.

        Returns:
            max_1_cone: Largest cone tangent to one atom
        """
        # Get the largest cone
        atoms = self._atoms
        alphas: List[float] = []
        for atom in atoms:
            alphas.append(atom.cone.angle)
        idx = int(np.argmax(alphas))
        max_1_cone = atoms[idx].cone

        # Check if all atoms are contained in cone. If yes, return cone,
        # otherwise, return None.
        in_atoms = []
        test_atoms = [atom for atom in atoms if atom not in max_1_cone.atoms]
        for atom in test_atoms:
            in_atoms.append(max_1_cone.is_inside(atom))
        if all(in_atoms):
            return max_1_cone
        else:
            return None

    def _search_two_cones(self) -> Optional[Cone]:
        """Search over cones tangent to two atoms.

        Returns:
            max_2_cone: Largest cone tangent to two atoms
        """
        # Create two-atom cones
        loop_atoms = self._loop_atoms
        cones = []
        for atom_i, atom_j in itertools.combinations(loop_atoms, r=2):
            cone = _get_two_atom_cone(atom_i, atom_j)
            cones.append(cone)

        # Select largest two-atom cone
        angles = [cone.angle for cone in cones]
        idx = int(np.argmax(angles))
        max_2_cone = cones[idx]
        self._max_2_cone = max_2_cone

        # Check if all atoms are contained in cone. If yes, return cone,
        # otherwise, return None
        in_atoms = []
        for atom in loop_atoms:
            in_atoms.append(max_2_cone.is_inside(atom))

        if all(in_atoms):
            return max_2_cone
        else:
            return None

    def _search_three_cones(self) -> Cone:
        """Search over cones tangent to three atoms.

        Returns:
            min_3_cone: Smallest cone tangent to three atoms
        """
        # Create three-atom cones
        loop_atoms = self._loop_atoms
        cones = []
        for atom_i, atom_j, atom_k in itertools.combinations(loop_atoms, r=3):
            three_cones = _get_three_atom_cones(atom_i, atom_j, atom_k)
            cones.extend(three_cones)

        # Get upper and lower bound to apex angle
        upper_bound = self._get_upper_bound()
        lower_bound = self._max_2_cone.angle

        # Remove cones from consideration which are outside the bounds
        remove_cones = []
        for cone in cones:
            if cone.angle - lower_bound < -1e-5 or upper_bound - cone.angle < -1e-5:
                remove_cones.append(cone)

        for cone in reversed(remove_cones):
            cones.remove(cone)

        # Keep only cones that encompass all atoms
        keep_cones = []
        for cone in cones:
            in_atoms = []
            for atom in loop_atoms:
                in_atoms.append(cone.is_inside(atom))
            if all(in_atoms):
                keep_cones.append(cone)

        # Take the smallest cone that encompasses all atoms
        cone_angles = [cone.angle for cone in keep_cones]
        idx = int(np.argmin(cone_angles))
        min_3_cone = keep_cones[idx]

        return min_3_cone

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        atom_scale: float = 1,
        background_color: str = "white",
        cone_color: str = "steelblue",
        cone_opacity: float = 0.75,
    ) -> None:
        """Draw a 3D representation of the molecule with the cone.

        Args:
            atom_scale: Scaling factor for atom size
            background_color: Background color for plot
            cone_color: Cone color
            cone_opacity: Cone opacity
        """
        # Set up plotter
        p = BackgroundPlotter()
        p.set_background(background_color)

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            radius = atom.radius * atom_scale
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            p.add_mesh(sphere, color=color, opacity=1, name=str(atom.index))

        # Determine direction and extension of cone
        cone_angle = math.degrees(self._cone.angle)
        coordinates = np.array([atom.coordinates for atom in self._atoms])
        radii = np.array([atom.radius for atom in self._atoms])
        if cone_angle > 180:
            normal = -self._cone.normal
        else:
            normal = self._cone.normal
        projected = np.dot(normal, coordinates.T) + np.array(radii)

        max_extension = np.max(projected)
        if cone_angle > 180:
            max_extension += 1

        # Make the cone
        cone = get_drawing_cone(
            center=[0, 0, 0] + (max_extension * normal) / 2,
            direction=-normal,
            angle=cone_angle,
            height=max_extension,
            capping=False,
            resolution=100,
        )
        p.add_mesh(cone, opacity=cone_opacity, color=cone_color)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def _get_two_atom_cone(atom_i: Atom, atom_j: Atom) -> Cone:
    """Creates a cone tangent to two atoms.

    Args:
        atom_i: First tangent atom
        atom_j: Second tangent atom

    Returns:
        cone: Cone tangent to the two atoms
    """
    # Get the cone angle
    cone_i = atom_i.cone
    cone_j = atom_j.cone
    beta_i = cone_i.angle
    beta_j = cone_j.angle
    beta_ij = math.acos(np.dot(atom_i.cone.normal, atom_j.cone.normal))
    alpha_ij = (beta_ij + beta_i + beta_j) / 2

    # Get the cone normal
    a_ij = (1 / math.sin(beta_ij)) * math.sin(0.5 * (beta_ij + beta_i - beta_j))
    b_ij = (1 / math.sin(beta_ij)) * math.sin(0.5 * (beta_ij - beta_i + beta_j))
    c_ij = 0

    n = a_ij * cone_i.normal + b_ij * cone_j.normal + c_ij
    n = n / np.linalg.norm(n)

    # Create cone
    angle = alpha_ij
    normal = n
    cone = Cone(angle, [atom_i, atom_j], normal)

    return cone


def _get_three_atom_cones(atom_i: Atom, atom_j: Atom, atom_k: Atom) -> List[Cone]:
    """Creates cones tangent to three atoms.

    Args:
        atom_i: First tangent atom
        atom_j: Second tangent atom
        atom_k: Third tangent atom

    Returns:
        cones: Cones tangent to the three atoms
    """
    # Set up vertex angles
    beta_i = atom_i.cone.angle
    beta_j = atom_j.cone.angle
    beta_k = atom_k.cone.angle

    # Set up angles between atom vectors
    beta_ij = math.acos(np.dot(atom_i.cone.normal, atom_j.cone.normal))

    # Set up normal vectors to atoms
    m_i = atom_i.cone.normal
    m_j = atom_j.cone.normal
    m_k = atom_k.cone.normal

    # Setup matrices
    u = np.array([math.cos(beta_i), math.cos(beta_j), math.cos(beta_k)])
    v = np.array([math.sin(beta_i), math.sin(beta_j), math.sin(beta_k)])
    N = np.array([np.cross(m_j, m_k), np.cross(m_k, m_i), np.cross(m_i, m_j)]).T
    P = N.T @ N
    gamma = np.dot(m_i, np.cross(m_j, m_k))

    # Set up coefficients of quadratic equation
    A = u @ P @ u
    B = v.T @ P @ v
    C = u.T @ P @ v
    D = gamma ** 2

    # Solve quadratic equation
    p2 = (A - B) ** 2 + 4 * C ** 2
    p1 = 2 * (A - B) * (A + B - 2 * D)
    p0 = (A + B - 2 * D) ** 2 - 4 * C ** 2

    roots = np.roots([p2, p1, p0])
    roots = np.real_if_close(roots, tol=1e10)
    roots[np.isclose(roots, 1, rtol=1e-9, atol=0.0)] = 1
    roots[np.isclose(roots, -1, rtol=1e-9, atol=0.0)] = -1

    cos_roots = [
        math.acos(roots[0]),
        2 * np.pi - math.acos(roots[0]),
        math.acos(roots[1]),
        2 * np.pi - math.acos(roots[1]),
    ]

    # Test roots and keep only those that are physical
    angles = []
    D_tests = []
    for root in cos_roots:
        alpha = root / 2
        test = (
            A * math.cos(alpha) ** 2
            + B * math.sin(alpha) ** 2
            + 2 * C * math.sin(alpha) * math.cos(alpha)
        )
        D_test = abs(test - D)
        angles.append(alpha)
        D_tests.append(D_test)
    angles = np.array(angles)
    D_tests = np.array(D_tests)
    physical_angles = angles[np.argsort(D_tests)][:2]

    # Create cones for physical angles
    cones = []
    for alpha in physical_angles:
        # Calculate normal vector
        a_ij = (
            math.cos(alpha - beta_i) - math.cos(alpha - beta_j) * math.cos(beta_ij)
        ) / math.sin(beta_ij) ** 2
        b_ij = (
            math.cos(alpha - beta_j) - math.cos(alpha - beta_i) * math.cos(beta_ij)
        ) / math.sin(beta_ij) ** 2
        c_ij_squared = 1 - a_ij ** 2 - b_ij ** 2 - 2 * a_ij * b_ij * math.cos(beta_ij)
        # Set c_ij_squared to 0 if negative due to numerical precision.
        if c_ij_squared < 0:
            c_ij_squared = 0
        c_ij = math.sqrt(c_ij_squared)
        p = N @ (u * math.cos(alpha) + v * math.sin(alpha)).reshape(-1)
        sign = np.sign(gamma) * np.sign(np.dot(p, np.cross(m_i, m_j)))
        if np.sign(c_ij) != sign:
            c_ij = -c_ij
        n = a_ij * m_i + b_ij * m_j + c_ij * 1 / math.sin(beta_ij) * np.cross(m_i, m_j)

        # Create cone
        cone = Cone(alpha, [atom_i, atom_j, atom_k], n)
        cones.append(cone)

    return cones


def cli(file: str) -> Any:
    """CLI for cone angle.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(ConeAngle, elements, coordinates)



"""Local force constant code."""

import functools
from os import PathLike
from typing import Any, Dict, FrozenSet, Iterable, List, Optional, Sequence, Union

import numpy as np

from morfeus.data import (
    AFU,
    AMU,
    ANGSTROM,
    ANGSTROM_TO_BOHR,
    atomic_masses,
    BOHR,
    BOHR_TO_ANGSTROM,
    C,
    DYNE,
    HARTREE,
)
from morfeus.geometry import (
    Angle,
    Bond,
    Dihedral,
    InternalCoordinates,
    kabsch_rotation_matrix,
)
from morfeus.io import read_geometry
from morfeus.typing import Array1D, Array2D, ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements


class LocalForce:
    """Calculates and stores the results from local force constant calculations.

    The method is described by Cremer in
    10.1002/(SICI)1097-461X(1998)67:1<1::AID-QUA1>3.0.CO;2-Z. Alternatively, the
    compliance matrix method can be used according to 10.1063/1.3413528

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)

    Attributes:
        internal_coordinates: Internal coordinates
        local_force_constants: Local force constants (mDyne/Å)
        local_frequencies: Local mode frequencies (cm⁻¹)
        n_imag: Number of normal modes with imaginary frequencies
    """

    internal_coordinates: List[Union[Bond, Angle, Dihedral]]
    local_force_constants: Array1D
    local_frequencies: Array1D
    n_imag: int
    _B_inv: Array2D
    _B: Array2D
    _coordinates: Array2D
    _D_full: Array2D
    _D: Array2D
    _elements: List[int]
    _fc_matrix: Array2D
    _force_constants: Array1D
    _ifc_matrix: Array2D
    _input_coordinates: Array2D
    _internal_coordinates: InternalCoordinates
    _masses: Array1D
    _normal_modes: Array2D
    _standard_coordinates: Array2D

    def __init__(
        self,
        elements: Optional[Union[Iterable[int], Iterable[str]]] = None,
        coordinates: Optional[ArrayLike2D] = None,
    ) -> None:
        # Set up attributes
        self._internal_coordinates = InternalCoordinates()
        self.internal_coordinates = self._internal_coordinates.internal_coordinates

        if elements is not None:
            elements = convert_elements(elements, output="numbers")
            self._elements = elements
            self._masses = np.array([atomic_masses[i] for i in elements])

        if coordinates is not None:
            self._coordinates = np.array(coordinates)

    def add_internal_coordinate(self, atoms: Sequence[int]) -> "LocalForce":
        """Add internal coordinate.

        Composed of two (bond), three (angle) or four atoms (dihedral).

        Args:
            atoms: Atom indices of internal coordinate

        Returns:
            self: Self
        """
        self._internal_coordinates.add_internal_coordinate(atoms)

        return self

    def compute_compliance(self) -> "LocalForce":
        """Compute local force constants with the compliance matrix method."""
        # Compute B matrix if it does not exists
        if not hasattr(self, "_B"):
            self._B = self._internal_coordinates.get_B_matrix(self._coordinates)

        # Compute compliance matrix and get local force constants
        C = self._B @ np.linalg.pinv(self._fc_matrix) @ self._B.T
        k_s = 1 / np.diag(C) * AFU / BOHR / (DYNE / 1000) * ANGSTROM

        self.local_force_constants = k_s

        return self

    def compute_frequencies(self) -> "LocalForce":
        """Compute local frequencies."""
        # Compute local frequencies
        M = np.diag(np.repeat(self._masses, 3))
        G = self._B @ np.linalg.inv(M) @ self._B.T
        frequencies = (
            np.sqrt(
                (
                    np.diag(G)
                    / AMU
                    * (self.local_force_constants / ANGSTROM * DYNE / 1000)
                )
                / (4 * np.pi ** 2 * C ** 2)
            )
            / 100
        )

        self.local_frequencies = frequencies

        return self

    def compute_local(  # noqa: C901
        self, project_imag: bool = True, cutoff: float = 1e-3
    ) -> "LocalForce":
        """Compute local force constants with the local modes approach.

        Args:
            project_imag: Whether to project out imaginary frequencies
            cutoff: Cutoff for low force constant (mDyne/Å)

        Returns:
            self: Self
        """
        # Compute D matrix from normal modes and B matrix
        if not hasattr(self, "_D"):
            if not hasattr(self, "_B"):
                self._B = self._internal_coordinates.get_B_matrix(self._coordinates)
            self._D = self._B @ self._normal_modes.T

        # Project out imaginary modes
        if project_imag is True and self.n_imag > 0:
            # Save full D matrix and force constant vector
            self._D_full = self._D

            # Remove force constants with imaginary force constants
            self._force_constants = self._force_constants[self.n_imag :]
            self._D = self._D[:, self.n_imag :]

        # Add cutoff to modes with low force constants
        if cutoff is not None:
            indices_small = np.where(np.array(self._force_constants) < cutoff)[0]
            if len(indices_small) > 0:
                self._force_constants[indices_small] = np.array(cutoff).reshape(-1)

        # Compute local mode force constants
        K = np.diag(self._force_constants)

        k_s = []
        for row in self._D:
            if not np.all(np.isclose(row, 0)):
                k = 1 / (row @ np.linalg.inv(K) @ np.conj(row).T)
                k_s.append(k)
            else:
                k_s.append(0.0)
        k_s = np.array(k_s)

        # Scale force constants due to projection of imaginary normal modes
        if project_imag and self.n_imag > 0:
            lengths = np.linalg.norm(self._D, axis=1)
            lengths_full = np.linalg.norm(self._D_full, axis=1)
            k_s *= lengths / lengths_full

        self.local_force_constants = k_s

        return self

    def detect_bonds(
        self,
        radii: Optional[ArrayLike1D] = None,
        radii_type: str = "pyykko",
        scale_factor: float = 1.2,
    ) -> "LocalForce":
        """Detect bonds based on scaled sum of covalent radii.

        Args:
            radii: Covalent radii (Å)
            radii_type: Covalent radii type: 'pyykko'
            scale_factor: Scale factor for covalent radii

        Returns:
            self: Self
        """
        self._internal_coordinates.detect_bonds(
            self._coordinates,
            self._elements,
            radii=radii,
            radii_type=radii_type,
            scale_factor=scale_factor,
        )

        return self

    def get_local_force_constant(self, atoms: Sequence[int]) -> float:
        """Return the local force constant between a set of atoms.

        Args:
            atoms: Atom indices of the internal coordinate

        Returns:
            force_constant: Local force constant (mDyne/Å, mDyne Å/rad²)

        Raises:
            ValueError: When no internal coordinate found
        """
        coordinate = _get_internal_coordinate(atoms)
        index = self.internal_coordinates.index(coordinate)
        if index is None:
            raise ValueError(f"No internal coordinate with these atoms: {atoms}")
        force_constant: float = self.local_force_constants[index]

        return force_constant

    def get_local_frequency(self, atoms: Sequence[int]) -> float:
        """Return the local frequency between a set of atoms.

        Args:
            atoms: Atom indices in the internal coordinate

        Returns:
            frequency: Local frequency (cm⁻¹)

        Raises:
            ValueError: When no internal coordinate found
        """
        coordinate = _get_internal_coordinate(atoms)
        index = self.internal_coordinates.index(coordinate)
        if index is None:
            raise ValueError(f"No internal coordinate with these atoms: {atoms}.")
        frequency: float = self.local_frequencies[index]

        return frequency

    def load_file(
        self, file: Union[str, PathLike], program: str, filetype: str
    ) -> "LocalForce":
        """Load data from external file.

        Args:
            file: File
            program: Program used to generate file: 'gaussian', 'unimovib' or 'xtb'
            filetype: Filetype. For 'gaussian': 'fchk' or 'log'. For 'unimovib': 'local'
                or 'log'. For 'xtb': 'hessian' or 'normal_modes'

        Returns:
            self: Self
        """
        choices = {
            "gaussian": {
                "fchk": self._parse_gaussian_fchk,
                "log": self._parse_gaussian_log,
            },
            "xtb": {
                "hessian": self._parse_xtb_hessian,
            },
            "unimovib": {
                "local": self._parse_unimovib_local,
                "log": self._parse_unimovib_log,
                "umv": self._parse_unimovib_umv,
            },
        }
        choices[program][filetype](file)

        return self

    def normal_mode_analysis(
        self,
        hessian: Optional[ArrayLike2D] = None,
        save_hessian: bool = False,
    ) -> "LocalForce":
        """Perform normal mode analysis.

        With projection of translations and vibrations to get normal modes and force
        constants.

        Args:
            hessian: User-supplied Hessian
            save_hessian: Save projected Hessian for use in compliance matrix method.

        Returns:
            self: Self
        """
        # Set up
        coordinates = np.array(self._coordinates) * ANGSTROM_TO_BOHR
        masses = self._masses
        if hessian is None:
            hessian = self._fc_matrix
        else:
            hessian = np.array(hessian)
        n_atoms = len(coordinates)

        # Create mass matrices
        M_minus = np.diag(np.repeat(masses, 3) ** (-1 / 2))
        M_plus = np.diag(np.repeat(masses, 3) ** (1 / 2))
        m_plus = np.repeat(masses, 3) ** (1 / 2)
        m_minus = np.repeat(masses, 3) ** (-1 / 2)

        # Mass-weight Hessian
        hessian_mw = M_minus @ hessian @ M_minus

        # Find center of mass
        com = np.sum(masses.reshape(-1, 1) * coordinates, axis=0) / np.sum(masses)

        # Shift origin to center of mass
        coordinates -= com

        # Construct translation vectors
        t_x = (np.tile(np.array([1, 0, 0]), n_atoms)).reshape(-1, 3)
        t_y = (np.tile(np.array([0, 1, 0]), n_atoms)).reshape(-1, 3)
        t_z = (np.tile(np.array([0, 0, 1]), n_atoms)).reshape(-1, 3)

        # Construct mass-weighted rotation vectors
        R_x = np.cross(coordinates, t_x).flatten() * m_plus
        R_y = np.cross(coordinates, t_y).flatten() * m_plus
        R_z = np.cross(coordinates, t_z).flatten() * m_plus

        # Mass-weight translation vectors
        T_x = t_x.flatten() * m_plus
        T_y = t_y.flatten() * m_plus
        T_z = t_z.flatten() * m_plus

        # Remove linear dependencies from translation/rotation space
        TR_vectors = np.vstack([T_x, T_y, T_z, R_x, R_y, R_z])
        Q, R = np.linalg.qr(TR_vectors.T)
        keep_indices = ~np.isclose(np.diag(R), 0, atol=1e-6, rtol=0)
        TR_vectors = Q.T[keep_indices]
        n_tr = len(TR_vectors)

        # Construct P matrix
        P = np.identity(n_atoms * 3)
        for vector in TR_vectors:
            P -= np.outer(vector, vector)

        # Project out translations and rotations
        hessian_proj = P.T @ hessian_mw @ P

        # Diagonalize
        eigenvalues, eigenvectors = np.linalg.eigh(hessian_proj)
        eigenvalues = eigenvalues[n_tr:]
        eigenvectors = eigenvectors[:, n_tr:]

        # Calculate cartesian displacements
        cart = eigenvectors.T * m_minus
        N = 1 / np.linalg.norm(cart, axis=1)
        norm_cart = cart * N.reshape(-1, 1)
        reduced_masses = N ** 2

        # Calculate frequencies and force constants
        n_imag = np.sum(eigenvalues < 0)
        frequencies = (
            np.sqrt(np.abs(eigenvalues) * HARTREE / BOHR ** 2 / AMU)
            / (2 * np.pi * C)
            / 100
        )
        frequencies[:n_imag] = -frequencies[:n_imag]
        force_constants = (
            4
            * np.pi ** 2
            * (frequencies * 100) ** 2
            * C ** 2
            * reduced_masses
            * AMU
            / (DYNE / 1000)
            * ANGSTROM
        )

        # Set up attributes
        self.n_imag = n_imag
        self._force_constants = force_constants
        self._normal_modes = norm_cart
        if save_hessian:
            self._fc_matrix = M_plus @ hessian_proj @ M_plus

        return self

    def print_report(
        self, angles: bool = False, dihedrals: bool = False, angle_units: bool = False
    ) -> None:
        """Print report of results.

        Args:
            angle_units: Wheter to convert angle and dihedral force constants to
                mDyne Å/rad²
            angles: Whether to print angles
            dihedrals: Whether to print dihedrals
        """
        # Print header
        if angle_units:
            unit = "mDyne/Å, mDyne Å/rad²"
        else:
            unit = "mDyne/Å"

        string = f"{'Coordinate':30s}" + f"{'Force constant ' + '(' + unit + ')':>50s}"
        if hasattr(self, "local_frequencies"):
            string += f"{'Frequency (cm⁻¹)':>30s}"
        print(string)

        # Print results for each internal
        sorted_coordinates = sorted(
            self.internal_coordinates, key=lambda x: (len(x.atoms), *x.atoms)
        )
        for coordinate in sorted_coordinates:
            # Check if internal coordinate is angle or dihedral
            if len(coordinate.atoms) == 3 and not angles:
                continue
            if len(coordinate.atoms) == 4 and not dihedrals:
                continue
            index = self.internal_coordinates.index(coordinate)
            force_constant = self.local_force_constants[index]

            # Convert units for angles and dihedrals
            if len(coordinate.atoms) > 2 and angle_units:
                force_constant = force_constant * BOHR_TO_ANGSTROM ** 2

            # Print out the results
            string = f"{repr(coordinate):30s}" + f"{force_constant:50.3f}"
            if hasattr(self, "local_frequencies"):
                frequency = self.local_frequencies[index]
                string += f"{frequency:30.0f}"
            print(string)

    def reset_internal_coordinates(self) -> "LocalForce":
        """Reset internal coordinate system."""
        self._internal_coordinates = InternalCoordinates()
        self.internal_coordinates = self._internal_coordinates.internal_coordinates

        return self

    def _parse_gaussian_fchk(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Read fchk file
        with open(file) as f:
            lines = f.readlines()

        # Set up read flags
        read_modes = False
        read_hessian = False
        read_vib_e2 = False
        read_ic = False
        read_atomic_numbers = False
        read_coordinates = False
        read_masses = False

        # Set up containers for reading data
        modes: List[float] = []
        hessian: List[float] = []
        vib_e2: List[float] = []
        internal_coordinates = []
        n_atoms: int
        n_imag: int
        atomic_numbers: List[int] = []
        masses: List[float] = []
        coordinates: List[float] = []

        # Parse fchk file
        for line in lines:
            # Read normal modes
            if read_modes:
                try:
                    split_line = line.strip().split()
                    modes.extend([float(value) for value in split_line])
                except ValueError:
                    read_modes = False
            # Read cartesian force constants
            elif read_hessian:
                try:
                    split_line = line.strip().split()
                    hessian.extend([float(value) for value in split_line])
                except ValueError:
                    read_hessian = False
            # Read normal mode force constants
            elif read_vib_e2:
                try:
                    split_line = line.strip().split()
                    vib_e2.extend([float(value) for value in split_line])
                except ValueError:
                    read_vib_e2 = False
            # Read internal coordinates
            elif read_ic:
                try:
                    split_line = line.strip().split()
                    internal_coordinates.extend([int(value) for value in split_line])
                except ValueError:
                    read_ic = False
            # Read atomic numbers
            elif read_atomic_numbers:
                try:
                    split_line = line.strip().split()
                    atomic_numbers.extend([int(value) for value in split_line])
                except ValueError:
                    read_atomic_numbers = False
            # Read atomic masses
            elif read_masses:
                try:
                    split_line = line.strip().split()
                    masses.extend([float(value) for value in split_line])
                except ValueError:
                    read_masses = False
            # Read coordinates
            elif read_coordinates:
                try:
                    split_line = line.strip().split()
                    coordinates.extend([float(value) for value in split_line])
                except ValueError:
                    read_coordinates = False
            # Read number of atoms
            if "Number of atoms" in line:
                n_atoms = int(line.strip().split()[4])
            # Read number of normal modes
            elif "Number of Normal Modes" in line:
                n_modes = int(line.strip().split()[5])
            # Read number of internal coordinates
            elif "Redundant internal coordinates" in line:
                n_redundant = int(line.strip().split()[5])
            elif "NImag" in line:
                n_imag = int(line.strip().split()[2])
            # Detect when to read data
            elif "Vib-Modes" in line:
                read_modes = True
            elif "Vib-AtMass" in line:
                read_masses = True
            elif "Cartesian Force Constants" in line:
                read_hessian = True
            elif "Vib-E2 " in line:
                read_vib_e2 = True
            elif "Redundant internal coordinate indices" in line:
                read_ic = True
            elif "Atomic numbers" in line:
                read_atomic_numbers = True
            elif "Current cartesian coordinates " in line:
                read_coordinates = True
        # Take out normal mode force constants
        force_constants = np.array(vib_e2[n_modes * 2 : n_modes * 3])

        # Construct force constant matrix from lower triangular matrix
        fc_matrix = np.zeros((n_atoms * 3, n_atoms * 3))
        fc_matrix[np.tril_indices_from(fc_matrix)] = hessian
        fc_matrix = np.triu(fc_matrix.T, 1) + fc_matrix

        # Take out the internal coordinates
        internal_coordinates = np.array(internal_coordinates)
        for i, coordinate in enumerate(np.split(internal_coordinates, n_redundant)):
            if all(coordinate >= 0):  # Sort out linear bends
                atoms = [i for i in coordinate if i != 0]
                self.add_internal_coordinate(atoms)

        # Convert coordinates to right Ångström
        coordinates = np.array(coordinates).reshape(-1, 3) * BOHR_TO_ANGSTROM

        # Set up attributes
        self._fc_matrix = fc_matrix
        self._normal_modes = np.array(modes).reshape(n_modes, n_atoms * 3)
        self._force_constants = force_constants
        self.n_imag = n_imag
        self._elements = convert_elements(atomic_numbers, output="numbers")
        self._coordinates = coordinates
        self._masses = np.array(masses)

    def _parse_gaussian_log(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Read the log file
        with open(file) as f:
            lines = f.readlines()

        # Set up read flags
        read_b_atoms = False
        read_b_vectors = False
        read_fc_matrix = False
        read_ifc_matrix = False
        read_input_orientation = False
        read_standard_orientation = False
        read_internal = False
        read_internal_modes = False
        read_hp_modes = False
        read_masses = False

        # Set up containers for reading data
        B_atom_map: Dict[int, List[int]] = {}
        B_vectors: Dict[int, List[float]] = {}
        normal_modes: List[List[List[float]]] = []
        internal_modes: List[List[float]] = []
        force_constants: List[float] = []
        masses: List[float] = []
        fc_matrix = np.array([])
        ifc_matrix = np.array([])
        input_coordinates: List[float] = []
        standard_coordinates: List[float] = []
        n_imag: int = 0
        n_atoms: int = 0
        internal_indices: Dict[FrozenSet[int], int] = {}
        atomic_numbers: List[int] = []
        coordinates: List[List[float]] = []

        # Parse through log file content
        counter = 0
        internal_names: List[str] = []
        internal_vector: List[float] = []

        values: Any
        value: Any
        for line in lines:
            # Read internal coordinate definitions
            if read_internal:
                if counter > 1:
                    if (
                        " --------------------------------------------------------------------------------"  # noqa: B950
                        in line
                    ):
                        read_internal = False
                        n_internals = len(internal_indices.items())
                    else:
                        split_line = line.strip().split()
                        name = split_line[1]
                        internal_names.append(name)
                        atoms = split_line[2][2:].replace(")", "").split(",")
                        atoms = [int(atom) for atom in atoms]
                        internal_indices[frozenset(atoms)] = counter - 2
                counter += 1
            # Read Cartesian force constant matrix
            elif read_fc_matrix:
                if " FormGI is forming" in line:
                    read_fc_matrix = False
                    fc_matrix = np.triu(fc_matrix.T, 1) + fc_matrix
                elif "          " in line:
                    column_indices = [int(value) for value in line.strip().split()]
                elif "D" in line:
                    split_line = line.strip().split()
                    row_index = int(split_line[0]) - 1
                    values = [
                        float(value.replace("D", "E")) for value in split_line[1:]
                    ]
                    for i, value in enumerate(values):
                        column_index = column_indices[i] - 1
                        fc_matrix[row_index, column_index] = value
            # Read internal force constant matrix
            elif read_ifc_matrix:
                if "Leave Link  716" in line:
                    read_ifc_matrix = False
                    ifc_matrix = np.triu(ifc_matrix.T, 1) + ifc_matrix
                elif "          " in line:
                    column_indices = [int(value) for value in line.strip().split()]
                elif "D" in line:
                    split_line = line.strip().split()
                    row_index = int(split_line[0]) - 1
                    values = [
                        float(value.replace("D", "E")) for value in split_line[1:]
                    ]
                    for i, value in enumerate(values):
                        column_index = column_indices[i] - 1
                        ifc_matrix[row_index, column_index] = value
            # Read atoms for creation of B matrix
            elif read_b_atoms:
                if " B Matrix in FormBX:" in line:
                    read_b_atoms = False
                else:
                    if counter == 0:
                        atoms = [int(value) for value in line.strip().split()]
                        for atom in atoms:
                            B_atom_map[atom] = []
                    if counter > 0 and counter < 5:
                        values = [int(value) for value in line.strip().split()[1:]]
                        for atom, value in zip(atoms, values):
                            B_atom_map[atom].append(value)
                    counter += 1
                    if counter == 5:
                        counter = 0
            # Read values of B matrix
            elif read_b_vectors:
                if (
                    " IB Matrix in Red2BG:" in line
                    or "Iteration" in line
                    or " G Matrix:" in line
                ):
                    read_b_vectors = False
                else:
                    if counter == 0:
                        atoms = [int(value) for value in line.strip().split()]
                        for atom in atoms:
                            B_vectors[atom] = []
                    if counter > 0 and counter < 13:
                        values = [
                            float(value.replace("D", "E"))
                            for value in line.strip().split()[1:]
                        ]
                        for atom, value in zip(atoms, values):
                            B_vectors[atom].append(value)
                    counter += 1
                    if counter == 13:
                        counter = 0
            # Read atomic coordinates in input orientation
            elif read_input_orientation:
                if counter > 3:
                    if (
                        "---------------------------------------------------------------------"  # noqa: B950
                        in line
                    ):
                        read_input_orientation = False
                    else:
                        strip_line = line.strip().split()
                        values = [float(value) for value in strip_line[3:]]
                        input_coordinates.append(values)
                        atomic_numbers.append(int(strip_line[1]))
                counter += 1
            # Read atomic coordinates in standard orientation
            elif read_standard_orientation:
                if counter > 3:
                    if (
                        "---------------------------------------------------------------------"  # noqa: B950
                        in line
                    ):
                        read_standard_orientation = False
                    else:
                        strip_line = line.strip().split()
                        values = [float(value) for value in strip_line[3:]]
                        standard_coordinates.append(values)
                counter += 1
            # Read decomposition of normal modes in internal coordinates
            elif read_internal_modes:
                if counter > 3:
                    if (
                        "--------------------------------------------------------------------------------"  # noqa: B950
                        in line
                    ):
                        read_internal_modes = False
                        internal_modes.append(internal_vector)
                    else:
                        value = float(line.strip().split()[3])
                        internal_vector.append(value)
                counter += 1
            # Read high-precision normal modes
            elif read_hp_modes:
                if counter < n_atoms * 3:
                    strip_line = line.strip().split()
                    values = [float(value) for value in strip_line[3:]]
                    coordinates.append(values)
                if counter == n_atoms * 3:
                    normal_modes.append(coordinates)
                    read_hp_modes = False
                counter += 1
            # Read atomic masses
            elif read_masses:
                if "Molecular mass: " in line:
                    read_masses = False
                elif "and mass" in line:
                    masses.append(float(line.strip().split()[8]))
            # Read number of atoms
            if " NAtoms=" in line and not n_atoms:
                n_atoms = int(line.strip().split()[1])
            # Read normal mode force constants
            elif "Frc consts" in line:
                split_line = line.strip().split()
                values = [float(value) for value in split_line[3:]]
                force_constants.extend(values)
            # Read number of imaginary frequencies
            elif "imaginary frequencies (negative Signs)" in line:
                n_imag = int(line.strip().split()[1])
            # Detect when to read data
            elif (
                "Name  Definition              Value          Derivative Info." in line
            ):
                read_internal = True
                counter = 1
                internal_names = []
            elif "- Thermochemistry -" in line:
                read_masses = True
            elif " IB Matrix in FormBX:" in line:
                read_b_atoms = True
                counter = 0
            elif " B Matrix in FormBX:" in line:
                read_b_vectors = True
                counter = 0
            elif " Force constants in Cartesian coordinates: " in line:
                read_fc_matrix = True
                fc_matrix = np.zeros((3 * n_atoms, 3 * n_atoms))
                counter = 0
            elif " Force constants in internal coordinates: " in line:
                read_ifc_matrix = True
                ifc_matrix = np.zeros((n_internals, n_internals))
                counter = 0
            elif "Input orientation: " in line:
                read_input_orientation = True
                counter = 0
            elif "Standard orientation: " in line:
                read_standard_orientation = True
                counter = 0
            elif "Normal Mode" in line:
                read_internal_modes = True
                counter = 1
                internal_vector = []
            elif " Coord Atom Element:" in line:
                read_hp_modes = True
                coordinates = []
                counter = 0

        # Process internal coordinates
        if len(internal_indices) > 0:
            for name, indices in zip(internal_names, internal_indices):
                if name[0] == "R" and len(indices) == 2:
                    self._internal_coordinates.add_internal_coordinate(list(indices))
                if name[0] == "A" and len(indices) == 3:
                    self._internal_coordinates.add_internal_coordinate(list(indices))
                if name[0] == "D" and len(indices) == 4:
                    self._internal_coordinates.add_internal_coordinate(list(indices))

        # Construct the B matrix from atoms and vectors
        if B_vectors:
            n_cartesian = n_atoms * 3
            B = np.zeros((n_internals, n_cartesian))
            for i in range(n_internals):
                for j, atom in enumerate(B_atom_map[i + 1]):
                    if atom:
                        B[i][(atom - 1) * 3 : (atom - 1) * 3 + 3] = B_vectors[i + 1][
                            j * 3 : j * 3 + 3
                        ]
            B_inv = np.linalg.pinv(B)
            # Detect whether the internal coordinate system is redundant
            if B.shape[0] == len(force_constants):
                self._redundant = False
            else:
                self._redundant = True
        else:
            B = np.array([])
            B_inv = np.array([])

        # Detect whether the input coordinates have been rotated. If so, rotate
        # B matrix and its inverse.
        input_coordinates = np.array(input_coordinates).reshape(-1, 3)
        standard_coordinates = np.array(standard_coordinates).reshape(-1, 3)

        if (
            not np.array_equal(input_coordinates, standard_coordinates)
            and standard_coordinates.size > 0
        ):
            rotation_i_to_s = kabsch_rotation_matrix(
                input_coordinates, standard_coordinates
            )
            if len(B) > 0:
                B = (rotation_i_to_s @ B.reshape(-1, 3).T).T.reshape(n_internals, -1)
                B_inv = np.linalg.pinv(B)

        # Set up attributes
        self._B = B
        self._B_inv = B_inv
        self._fc_matrix = fc_matrix
        if fc_matrix.size == 0 and ifc_matrix.size > 0:
            self._fc_matrix = B.T @ ifc_matrix @ B
        self._ifc_matrix = ifc_matrix
        self._force_constants = np.array(force_constants)
        if len(normal_modes) > 0:
            self._normal_modes = np.hstack(normal_modes).T
        else:
            self._normal_modes = np.array([])
        if len(internal_modes) > 0:
            self._D = np.vstack(internal_modes).T
        self.n_imag = n_imag
        self._masses = np.array(masses)
        self._input_coordinates = input_coordinates
        self._standard_coordinates = standard_coordinates
        self._coordinates = input_coordinates
        self._elements = convert_elements(atomic_numbers, output="numbers")

    def _parse_unimovib_local(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Read file
        with open(file) as f:
            lines = f.readlines()

        # Set up flags for reading
        read_masses = False
        read_atomic_numbers = False
        read_coordinates = False
        read_hessian = False
        read_normal_modes = False

        # Set up containers for data
        masses = []
        atomic_numbers = []
        coordinates = []
        hessian = []
        normal_modes = []

        # Parse file
        for line in lines:
            # Read atomic masses
            if read_masses:
                if " $ZA  $END" in line:
                    read_masses = False
                else:
                    split_line = line.strip().split()
                    masses.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read atomic numbers
            if read_atomic_numbers:
                if " $XYZ  $END" in line:
                    read_atomic_numbers = False
                else:
                    split_line = line.strip().split()
                    atomic_numbers.extend(
                        [int(float(value.replace("D", "E"))) for value in split_line]
                    )
            # Read coordinates
            if read_coordinates:
                if " $FFX  $END" in line:
                    read_coordinates = False
                else:
                    split_line = line.strip().split()
                    coordinates.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read Hessian
            if read_hessian:
                if " $NMMODE  $END" in line:
                    read_hessian = False
                else:
                    split_line = line.strip().split()
                    hessian.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read normal modes
            if read_normal_modes:
                if " $APT" in line:
                    read_normal_modes = False
                else:
                    split_line = line.strip().split()
                    normal_modes.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read number of atoms and normal modes
            if " $CONTRL" in line:
                split_line = line.strip().split()
                n_atoms = int(split_line[1].split("=")[1])
                n_modes = int(split_line[2].split("=")[1])
            # Detect when to read data
            if " $AMASS  $END" in line:
                read_masses = True
            if " $ZA  $END" in line:
                read_atomic_numbers = True
            if " $XYZ  $END" in line:
                read_coordinates = True
            if " $FFX  $END" in line:
                read_hessian = True
            if " $NMMODE  $END" in line:
                read_normal_modes = True

        # Convert data to right format
        coordinates = np.array(coordinates).reshape(-1, 3) * BOHR_TO_ANGSTROM
        hessian = np.array(hessian).reshape(n_atoms * 3, n_atoms * 3)
        normal_modes = np.array(normal_modes).reshape(-1, n_atoms * 3)[:n_modes]
        elements = convert_elements(atomic_numbers, output="numbers")

        # Set up attributes
        self._normal_modes = normal_modes
        self._masses = np.array(masses)
        self._fc_matrix = hessian
        self._coordinates = coordinates
        self._elements = elements

    def _parse_unimovib_log(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Read file
        with open(file) as f:
            lines = f.readlines()

        # Set up read flags
        read_coordinates = False
        read_vibrations = False
        read_normal_modes = False

        # Set up data containers
        atomic_numbers = []
        masses = []
        coordinates = []
        force_constants = []
        normal_modes = []
        frequencies = []

        # Parse file
        normal_modes_chunk: List[List[float]] = []
        counter = 0
        n_modes_chunk = 0
        values: Any
        for line in lines:
            # Read coordinates, atomic numbers and atomic masses
            if read_coordinates:
                if counter > 0:
                    if (
                        " ------------------------------------------------------------------------------------------"  # noqa: B950
                        in line
                    ):
                        read_coordinates = False
                    else:
                        split_line = line.strip().split()
                        atomic_numbers.append(int(split_line[2]))
                        coordinates.extend([float(value) for value in split_line[3:6]])
                        masses.append(float(split_line[6]))
                counter += 1
            # Read normal modes and force constants
            if read_vibrations:
                if " Results of translations and rotations:" in line:
                    read_vibrations = False
                if read_normal_modes:
                    split_line = line.strip().split()
                    if len(split_line) == 0:
                        read_normal_modes = False
                        normal_modes.extend(normal_modes_chunk)
                    else:
                        values = [float(value) for value in split_line[2:]]
                        for i in range(n_modes_chunk):
                            normal_modes_chunk[i].append(values[i * 3 : i * 3 + 3])
                elif "Irreps" in line:
                    split_line = line.strip().split()
                    n_modes_chunk = len(split_line) - 1
                    normal_modes_chunk = [[] for i in range(n_modes_chunk)]
                elif "Force constants" in line:
                    split_line = line.strip().split()
                    force_constants.extend([float(value) for value in split_line[2:]])
                elif "Frequencies" in line:
                    split_line = line.strip().split()
                    frequencies.extend([float(value) for value in split_line[2:]])
                elif "        Atom  ZA" in line:
                    read_normal_modes = True
            # Detect when to read data
            if "No.   Atom    ZA" in line:
                read_coordinates = True
                counter = 0
            if "Results of vibrations:" in line:
                read_vibrations = True

        # Convert quantities to right format
        n_atoms = len(masses)
        coordinates = np.array(coordinates).reshape(-1, 3)
        normal_modes = np.array(normal_modes).reshape(-1, n_atoms * 3)
        elements = convert_elements(atomic_numbers, output="numbers")
        force_constants = np.array(force_constants)
        frequencies = np.array(frequencies)
        masses = np.array(masses)

        # Detect imaginary modes
        self.n_imag = np.sum(frequencies < 0)

        # Set up attributes
        self._elements = elements
        self._coordinates = coordinates
        self._masses = masses
        self._force_constants = force_constants
        self._normal_modes = normal_modes

    def _parse_unimovib_umv(self, file: Union[str, PathLike]) -> None:  # noqa: C901
        # Read file
        with open(file) as f:
            lines = f.readlines()

        # Set up flags for reading
        read_n_atoms = False
        read_masses = False
        read_atomic_numbers = False
        read_coordinates = False
        read_hessian = False

        # Set up containers for data
        n_atoms: int
        masses = []
        atomic_numbers = []
        coordinates = []
        hessian = []

        # Parse file
        for line in lines:
            # Read number of atoms
            if read_n_atoms:
                if "AMASS" in line:
                    read_n_atoms = False
                else:
                    n_atoms = int(line.strip())
            # Read atomic masses
            if read_masses:
                if "ZA" in line:
                    read_masses = False
                else:
                    split_line = line.strip().split()
                    masses.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read atomic numbers
            if read_atomic_numbers:
                if "XYZ" in line:
                    read_atomic_numbers = False
                else:
                    split_line = line.strip().split()
                    atomic_numbers.extend(
                        [int(float(value.replace("D", "E"))) for value in split_line]
                    )
            # Read coordinates
            if read_coordinates:
                if "FFX" in line:
                    read_coordinates = False
                else:
                    split_line = line.strip().split()
                    coordinates.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read Hessian
            if read_hessian:
                if "APT" in line:
                    read_hessian = False
                else:
                    split_line = line.strip().split()
                    hessian.extend(
                        [float(value.replace("D", "E")) for value in split_line]
                    )
            # Read number of atoms and normal modes
            if "NATM" in line:
                read_n_atoms = True
            # Detect when to read data
            if "AMASS" in line:
                read_masses = True
            if "ZA" in line:
                read_atomic_numbers = True
            if "XYZ" in line:
                read_coordinates = True
            if "FFX" in line:
                read_hessian = True

        # Convert data to right format
        coordinates = np.array(coordinates).reshape(-1, 3) * BOHR_TO_ANGSTROM
        hessian = np.array(hessian).reshape(n_atoms * 3, n_atoms * 3)
        elements = convert_elements(atomic_numbers, output="numbers")

        # Set up attributes
        self._masses = np.array(masses)
        self._fc_matrix = hessian
        self._coordinates = coordinates
        self._elements = elements

    def _parse_xtb_hessian(self, file: Union[str, PathLike]) -> None:
        # Read hessian file
        with open(file) as f:
            lines = f.readlines()

        # Parse file
        hessian = []
        for line in lines:
            try:
                hessian.extend([float(value) for value in line.strip().split()])
            except ValueError:
                pass
        # Set up force constant matrix
        dimension = int(np.sqrt(len(hessian)))
        self._fc_matrix = np.array(hessian).reshape(dimension, dimension)

    def __repr__(self) -> str:
        n_internal = len(self.internal_coordinates)
        return f"{self.__class__.__name__}({n_internal!r} internal coordinates)"


def _get_internal_coordinate(atoms: Sequence[int]) -> Union[Bond, Angle, Dihedral]:
    """Returns internal coordinate."""
    # Return bond, angle or dihedral
    if len(atoms) == 2:
        return Bond(*atoms)
    elif len(atoms) == 3:
        return Angle(*atoms)
    elif len(atoms) == 4:
        return Dihedral(*atoms)
    else:
        raise ValueError(f"Sequence of atoms must be 2-4 atoms, not {len(atoms)}.")


def cli(file: Optional[str] = None) -> Any:
    """CLI for local force.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    if file is not None:
        elements, coordinates = read_geometry(file)
        return functools.partial(LocalForce, elements, coordinates)
    else:
        return functools.partial(LocalForce, None, None)



"""Interface to quantum-chemical programs."""

import typing
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np

from morfeus.data import ANGSTROM_TO_BOHR, BOHR_TO_ANGSTROM
from morfeus.typing import ArrayLike2D
from morfeus.utils import convert_elements, Import, requires_dependency

if typing.TYPE_CHECKING:
    import qcelemental as qcel
    import qcengine as qcng


@requires_dependency([Import(module="qcengine", alias="qcng")], globals())
def optimize_qc_engine(
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: ArrayLike2D,
    charge: Optional[int] = None,
    multiplicity: Optional[int] = None,
    connectivity_matrix: Optional[ArrayLike2D] = None,
    program: str = "xtb",
    model: Optional[Dict[str, Any]] = None,
    keywords: Optional[Dict[str, Any]] = None,
    local_options: Optional[Dict[str, Any]] = None,
    procedure: str = "berny",
    return_trajectory: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Optimize molecule with QCEngine.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        charge: Molecular charge
        multiplicity: Multiplicity
        connectivity_matrix: Connectivity matrix
        program: QCEngine program
        model: QCEngine model
        keywords: QCEngine keywords
        local_options: QCEngine local options
        procedure: QCEngine procedure
        return_trajectory: Return coordinates for all steps

    Returns:
        opt_coordinates (ndarray): Conformer coordinates (Å)
        energies (ndarray): Energies for all steps (a.u.)

    Raises:
        Exception: When QCEngine calculation fails
    """
    if (
        program.lower() == "rdkit"
        and charge is not None
        and connectivity_matrix is not None
    ):
        _check_qcng_rdkit(charge, connectivity_matrix)

    # Set defaults
    if model is None:
        model = {"method": "GFN2-xTB"}
    if keywords is None:
        keywords = {}
    if local_options is None:
        local_options = {}

    # Create molecule object
    molecule = _generate_qcel_molecule(
        elements, coordinates, charge, multiplicity, connectivity_matrix
    )

    # Create optimization input
    opt_input = {
        "keywords": {"program": program},
        "input_specification": {
            "driver": "gradient",
            "model": model,
            "keywords": keywords,
        },
        "initial_molecule": molecule,
    }

    # Perform optimization
    opt = qcng.compute_procedure(
        opt_input, procedure=procedure, local_options=local_options
    )
    if not opt.success:
        raise Exception(opt.error.error_message)

    # Take out results
    energies = np.array(opt.energies)
    if return_trajectory:
        opt_coordinates = np.array(
            [result.molecule.geometry for result in opt.trajectory]
        )
    else:
        opt_coordinates = opt.final_molecule.geometry
    opt_coordinates *= BOHR_TO_ANGSTROM

    return opt_coordinates, energies


@requires_dependency([Import(module="qcengine", alias="qcng")], globals())
def sp_qc_engine(
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: Sequence[Sequence[float]],
    charge: Optional[int] = None,
    multiplicity: Optional[int] = None,
    connectivity_matrix: Optional[Sequence[Sequence[int]]] = None,
    program: str = "xtb",
    model: Optional[Dict[str, Any]] = None,
    keywords: Optional[Dict[str, Any]] = None,
    local_options: Optional[Dict[str, Any]] = None,
) -> float:
    """Single-point calculation with QCEngine.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        charge: Molecular charge
        multiplicity: Molecular multiplicity
        connectivity_matrix: Connectivity matrix
        program: QCEngine program
        model: QCEngine model
        keywords: QCEngine keywords
        local_options: QCEngine local options

    Returns:
        energy: Energy (a.u.)

    Raises:
        Exception: When QCEngine calculation fails
    """
    if (
        program.lower() == "rdkit"
        and charge is not None
        and connectivity_matrix is not None
    ):
        _check_qcng_rdkit(charge, connectivity_matrix)

    # Set defaults
    if model is None:
        model = {"method": "GFN2-xTB"}
    if keywords is None:
        keywords = {}
    if local_options is None:
        local_options = {}

    # Crate molecule object
    molecule = _generate_qcel_molecule(
        elements, coordinates, charge, multiplicity, connectivity_matrix
    )

    # Create sp input
    sp_input = qcel.models.AtomicInput(
        molecule=molecule,
        driver="energy",
        model=model,
        keywords=keywords,
    )

    # Perform sp calculation
    sp = qcng.compute(sp_input, program=program, local_options=local_options)
    if not sp.success:
        raise Exception(sp.error.error_message)

    # Take out results
    energy: float = sp.return_result

    return energy


def _check_qcng_rdkit(charge: int, connectivity_matrix: ArrayLike2D) -> None:
    """Check qncg calculation for RDKit incompatibilities."""
    if charge != 0:
        raise Exception("QCEngine using RDKit does not work with charged molecules.")
    if np.any(~np.isin(connectivity_matrix, [0, 1, 2, 3])):
        raise Exception(
            "QCEngine using RDKit cannot handle bond orders different from "
            "1, 2 or 3."
        )


@requires_dependency([Import(module="qcelemental", alias="qcel")], globals())
def _generate_qcel_molecule(
    elements: Union[Iterable[int], Iterable[str]],
    coordinates: ArrayLike2D,
    charge: Optional[int] = None,
    multiplicity: Optional[int] = None,
    connectivity_matrix: Optional[ArrayLike2D] = None,
) -> "qcel.models.Molecule":
    """Generate QCElemental molecule object.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        charge: Molecular charge
        multiplicity: Molecular multiplicity
        connectivity_matrix: Connectivity matrix

    Returns:
        molecule: QCElemental molecule object.
    """
    # Generate bond order list from connectivity matrix
    bos: Optional[List[Tuple[int, int, int]]]
    if connectivity_matrix is not None:
        connectivity_matrix = np.array(connectivity_matrix)
        bos = []
        i, j = np.tril_indices_from(connectivity_matrix)
        for k, l in zip(i, j):
            if k != l:
                bo = int(connectivity_matrix[k, l])
                if bo != 0:
                    bos.append((k, l, bo))
    else:
        bos = None

    # Create molecule object
    elements = np.array(convert_elements(elements, output="symbols"))
    coordinates = np.array(coordinates) * ANGSTROM_TO_BOHR
    molecule = qcel.models.Molecule(
        symbols=elements,
        geometry=coordinates,
        molecular_charge=charge,
        connectivity=bos,
        molecular_multiplicity=multiplicity,
    )

    return molecule
